"""Checker for usage of disallowed schemas."""

from pglast import ast, visitors

from pgrubic.core import config, linter


class DisallowedSchema(linter.BaseChecker):
    """## **What it does**
    Checks for usage of disallowed schemas.

    ## **Why not?**
    If a schema has been included in the disallowed_schemas config, it is not allowed.

    ## **When should you?**
    Do you really want to use a disallowed schema?

    ## **Use instead:**
    Allowed schemas.

    ## **Configuration**
    `disallowed-schemas`: List of disallowed schemas.
    """

    is_auto_fixable: bool = True

    help: str = "Do not use disallowed schema"

    def visit_RangeVar(
        self,
        ancestors: visitors.Ancestor,
        node: ast.RangeVar,
    ) -> None:
        """Visit RangeVar."""
        # We are only interested in object creation
        if isinstance(
            abs(ancestors).node,
            ast.CreateStmt
            | ast.ViewStmt
            | ast.CompositeTypeStmt
            | ast.CreateSeqStmt
            | ast.IntoClause,
        ):
            for schema in self.config.lint.disallowed_schemas:
                if node.schemaname == schema.name:
                    self.violations.add(
                        linter.Violation(
                            rule_code=self.code,
                            rule_name=self.name,
                            rule_category=self.category,
                            line_number=self.line_number,
                            column_offset=self.column_offset,
                            line=self.line,
                            statement_location=self.statement_location,
                            description=f"Schema '{node.schemaname}' is disallowed in"
                            f" config with reason: '{schema.reason}'"
                            f", use '{schema.use_instead}' instead",
                            is_auto_fixable=self.is_auto_fixable,
                            is_fix_enabled=self.is_fix_enabled,
                            help=self.help,
                        ),
                    )

                    self._fix_range_var(node, schema)

    def _fix_range_var(
        self,
        node: ast.RangeVar,
        schema: config.DisallowedSchema,
    ) -> None:
        """Fix violation."""
        node.schemaname = schema.use_instead

    def visit_CreateEnumStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateEnumStmt,
    ) -> None:
        """Visit CreateEnumStmt."""
        schema_name: str | None = (
            node.typeName[0].sval if len(node.typeName) > 1 else None
        )

        for schema in self.config.lint.disallowed_schemas:
            if schema_name == schema.name:
                self.violations.add(
                    linter.Violation(
                        rule_code=self.code,
                        rule_name=self.name,
                        rule_category=self.category,
                        line_number=self.line_number,
                        column_offset=self.column_offset,
                        line=self.line,
                        statement_location=self.statement_location,
                        description=f"Schema '{schema_name}' is disallowed in"
                        f" config with reason: '{schema.reason}'"
                        f", use '{schema.use_instead}' instead",
                        is_auto_fixable=self.is_auto_fixable,
                        is_fix_enabled=self.is_fix_enabled,
                        help=self.help,
                    ),
                )

                self._fix_enum(node, schema)

    def _fix_enum(
        self,
        node: ast.CreateEnumStmt,
        schema: config.DisallowedSchema,
    ) -> None:
        """Fix violation."""
        node.typeName[0].sval = schema.use_instead

    def visit_CreateFunctionStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateFunctionStmt,
    ) -> None:
        """Visit CreateFunctionStmt."""
        schema_name: str | None = (
            node.funcname[0].sval if len(node.funcname) > 1 else None
        )

        for schema in self.config.lint.disallowed_schemas:
            if schema_name == schema.name:
                self.violations.add(
                    linter.Violation(
                        rule_code=self.code,
                        rule_name=self.name,
                        rule_category=self.category,
                        line_number=self.line_number,
                        column_offset=self.column_offset,
                        line=self.line,
                        statement_location=self.statement_location,
                        description=f"Schema '{schema_name}' is disallowed in"
                        f" config with reason: '{schema.reason}'"
                        f", use '{schema.use_instead}' instead",
                        is_auto_fixable=self.is_auto_fixable,
                        is_fix_enabled=self.is_fix_enabled,
                        help=self.help,
                    ),
                )

                self._fix_function(node, schema)

    def _fix_function(
        self,
        node: ast.CreateFunctionStmt,
        schema: config.DisallowedSchema,
    ) -> None:
        """Fix violation."""
        node.funcname[0].sval = schema.use_instead



"""Checker for creation of enum."""

from pglast import ast, visitors

from pgrubic.core import linter


class CreateEnum(linter.BaseChecker):
    """## **What it does**
    Checks for creation of enum.

    ## **Why not?**
    Enum types are primarily intended for static sets of values, though there is support
    for adding new values to an existing enum type, and for renaming values however
    existing values cannot be removed from an enum type, nor can the sort ordering of
    such values be changed, short of dropping and re-creating the enum type.

    ## **When should you?**
    If your set of values are fixed such that you won't be removing values
    in the future. Just don't use it automatically without thinking about it.

    ## **Use instead:**
    Mapping table.
    """

    def visit_CreateEnumStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateEnumStmt,
    ) -> None:
        """Visit CreateEnumStmt."""
        self.violations.add(
            linter.Violation(
                rule_code=self.code,
                rule_name=self.name,
                rule_category=self.category,
                line_number=self.line_number,
                column_offset=self.column_offset,
                line=self.line,
                statement_location=self.statement_location,
                description="Prefer mapping table to enum",
                is_auto_fixable=self.is_auto_fixable,
                is_fix_enabled=self.is_fix_enabled,
                help="Use mapping table",
            ),
        )



"""Test yaml test cases rules."""

import typing
import pathlib

import pytest
from pglast import parser

from tests import conftest
from pgrubic import core


class RuleTestCase(typing.NamedTuple):
    """Rule test case."""

    rule: str
    sql_fail: str | None
    sql_pass: str | None
    sql_fix: str | None


@pytest.mark.parametrize(
    ("rule", "test_id", "test_case"),
    conftest.load_test_cases(
        test_case_type=conftest.TestCaseType.RULE,
        directory=pathlib.Path("tests/fixtures/rules"),
    ),
)
def test_rules(
    linter: core.Linter,
    rule: str,
    test_id: str,
    test_case: dict[str, str],
) -> None:
    """Test rules."""
    parsed_test_case = RuleTestCase(
        rule=rule,
        sql_fail=test_case.get("sql_fail"),
        sql_pass=test_case.get("sql_pass"),
        sql_fix=test_case.get("sql_fix"),
    )

    config_overrides: dict[str, typing.Any] = typing.cast(
        dict[str, typing.Any],
        test_case.get("config", {}),
    )

    # Apply overrides to global configuration
    conftest.update_config(linter.config, config_overrides)

    if parsed_test_case.sql_fail:
        # Set fix flag
        linter.config.lint.fix = bool(parsed_test_case.sql_fix)

        linting_result = linter.run(
            source_file=f"{parsed_test_case.rule}.sql",
            source_code=parsed_test_case.sql_fail,
        )

        assert any(
            violation.rule_code == rule for violation in linting_result.violations
        ), f"Test failed: No violations found for rule: `{rule}` in `{test_id}`"

        if parsed_test_case.sql_fix:
            assert linting_result.fixed_sql == parsed_test_case.sql_fix

            # Check that the fixed sql is valid
            try:
                parser.parse_sql(linting_result.fixed_sql)
            except parser.ParseError as error:
                msg = f"Formatted code is not a valid syntax: {error!s}"
                raise ValueError(msg) from error

    if parsed_test_case.sql_pass:
        linting_result = linter.run(
            source_file=f"{parsed_test_case.rule}.sql",
            source_code=parsed_test_case.sql_pass,
        )

        assert not any(
            violation.rule_code == rule for violation in linting_result.violations
        ), f"""Test failed: Violations found for rule: `{rule}` in `{test_id}` which should pass"""  # noqa: E501



"""Checker for adding of auto increment identity column."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class AddingAutoIncrementIdentityColumn(linter.BaseChecker):
    """## **What it does**
    Checks adding of auto increment identity column.

    ## **Why not?**
    Adding an auto increment identity column to an already populated table will
    have to backfill the newly added column, causing the table to be locked
    in which no other operations can be performed on the table for the duration
    of the backfill. This will cause downtime if the table is concurrently
    being accessed by other clients.

    ## **When should you?**
    If the table is empty.
    If the table is not empty but is not being concurrently accessed.

    ## **Use instead:**
    1. Create a new column typed bigint, nullable.
    2. Create a sequence.
    3. Set the next value of the sequence to the total number of rows in the table
       with enough offset.
    4. Set the default value of the new column to the next value of the sequence.
    5. Backfill the new column for all existing rows.
    6. Add a check constraint: **CHECK (column IS NOT NULL) NOT VALID**.
    7. Validate the constraint.
    8. Set the column as NOT NULL
    9. Drop the constraint.
    10. Get the last value of the sequence, with enough offset.
    11. In a single transaction:
        - Drop the default on the new column.
        - Drop the sequence created in step 2.
        - Add GENERATED ALWAYS AS IDENTITY constraint instead,
        specifying the start option as the value from step 10.
    """

    def visit_Constraint(
        self,
        ancestors: visitors.Ancestor,
        node: ast.Constraint,
    ) -> None:
        """Visit Constraint."""
        alter_table_cmd: visitors.Ancestor = ancestors.find_nearest(ast.AlterTableCmd)

        if (
            alter_table_cmd
            and alter_table_cmd.node.subtype == enums.AlterTableType.AT_AddColumn
            and node.contype == enums.ConstrType.CONSTR_IDENTITY
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Adding auto increment identity column is not safe",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Split the operation into multiple steps",
                ),
            )



"""Checker for date columns without defined suffix, by default `_date`."""

from pglast import ast, visitors

from pgrubic.core import linter


class DateColumnWithoutSuffix(linter.BaseChecker):
    """## **What it does**
    Checks that date columns are suffixed with the defined suffix, by default `_date`.

    ## **Why not?**
    Adding `_date` to a date column name makes it clear that the value represents the
    date when something happened. For example, `created_date` indicates the date when a
    record was created, and `updated_date` indicates the date when a record was last
    updated.

    Date columns could easily be confused with other types of data if not clearly
    named. For example, a column named created might be unclear — does it represent a
    boolean flag, a date, or something else? `created_date` removes this ambiguity by
    specifying that it's a date.

    ## **When should you?**
    Almost Never.

    ## **Use instead:**
    Add the defined suffix or the default `_date` to to the date column name.

    ## **Configuration**
    `date-column-suffix`: Specify the suffix for date columns.
    """

    is_auto_fixable: bool = True

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if (
            node.typeName.names[-1].sval == "date"
            and node.colname
            and not node.colname.endswith(self.config.lint.date_column_suffix)
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Date column name should be suffixed with"
                    f" `{self.config.lint.date_column_suffix}`",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Add the set suffix to the date column",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.ColumnDef) -> None:
        """Fix violation."""
        node.colname += self.config.lint.date_column_suffix



"""Indexes movement to tablespace."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class IndexesMovementToTablespace(linter.BaseChecker):
    """## **What it does**
    Checks indexes movement to a different tablespace.

    ## **Why not?**
    Moving indexes to a different tablespace acquires an **ACCESS EXCLUSIVE** lock on the
    respective tables, blocking other accesses until the movement is completed.
    This will cause downtime if the tables are concurrently being accessed by other
    clients.

    ## **When should you?**
    If the tables are empty.
    If the tables are not empty but is not being concurrently accessed.

    ## **Use instead:**
    Have a look at pg_repack as an alternative.
    """

    def visit_AlterTableMoveAllStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.AlterTableMoveAllStmt,
    ) -> None:
        """Visit AlterTableMoveAllStmt."""
        if (
            node.objtype == enums.ObjectType.OBJECT_INDEX
            and node.new_tablespacename != node.orig_tablespacename
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Indexes movement to tablespace",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Have a look at pg_repack as an alternative",
                ),
            )



"""Checker for char."""

from pglast import ast, visitors

from pgrubic.core import linter


class Char(linter.BaseChecker):
    r"""## **What it does**
    Checks for usage of char.

    ## **Why not?**
    Any string you insert into a char(n) field will be padded with spaces to the
    declared width. That's probably not what you actually want.

    The manual says:

    > Values of type character are physically padded with spaces to the specified
    > width n, and are stored and displayed that way.
    > However, trailing spaces are treated as semantically insignificant and disregarded
    > when comparing two values of type character. In collations where whitespace is
    > significant, this behavior can produce unexpected results; for example
    > SELECT 'a '::CHAR(2) collate "C" < E'a\n'::CHAR(2) returns true, even though
    > C locale would consider a space to be greater than a newline.
    > Trailing spaces are removed when converting a character value to one of the
    > other string types. Note that trailing spaces are semantically significant in
    > character varying and text values, and when using pattern matching, that is LIKE
    > and regular expressions. That should scare you off it.

    The space-padding does waste space, but doesn't make operations on it any faster;
    in fact the reverse, thanks to the need to strip spaces in many contexts.

    It's important to note that from a storage point of view char(n) is not a
    fixed-width type. The actual number of bytes varies since characters may take more
    than one byte, and the stored values are therefore treated as variable-length anyway
    (even though the space padding is included in the storage).

    Sometimes people respond to "don't use char(n)" with "but my values must always be
    exactly N characters long" (e.g. country codes, hashes, or identifiers from some
    other system). It is still a bad idea to use char(n) even in these cases.

    Remember, there is no performance benefit whatsoever to using char(n).
    In fact the reverse is true. One particular problem that comes up is that if you try
    and compare a char(n) field against a parameter where the driver has explicitly
    specified a type of text or varchar, you may be unexpectedly unable to use an index
    for the comparison. This can be hard to debug since it doesn't show up on
    manual queries.

    ## **When should you?**
    When you're porting very, very old software that uses fixed width fields. Or when
    you read the snippet from the manual above and think "yes, that makes perfect sense
    and is a good match for my requirements" rather than gibbering and running away.

    ## **Use instead:**
    1. text
    2. text with a constraint that enforces a maximum string length
    """

    is_auto_fixable: bool = True

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if node.typeName.names[-1].sval in ["bpchar", "char"]:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Prefer text to char",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use text",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.ColumnDef) -> None:
        """Fix violation."""
        node.typeName = ast.TypeName(
            names=(
                {
                    "@": "String",
                    "sval": "text",
                },
            ),
        )
